/** \addtogroup tuePrvlkfTrkMgt
 * @{
 * \file        tue_prv_lkf_trackManagement.h
 *
 *
 *
 *
 *          (C) Copyright Tuerme Inc. All rights reserved.
 *
 */

#ifndef TUE_PRV_LKF_TRACKMANAGEMENT_H
#define TUE_PRV_LKF_TRACKMANAGEMENT_H

#ifdef __cplusplus
extern "C" {
#endif

/*==================[inclusions]============================================*/
#include "tue_prv_common_types.h"
#include "TueObjFusn_TrackableListType.h"
#include "TueObjFusn_ObjectListType.h"
#include "TueObjFusn_MatchIndex.h"
#include "TueObjFusn_EgoMotionType.h"

/*==================[macros]================================================*/
/*==================[type definitions]======================================*/

/*****************************************************************************
   FORWARD DECLARATIONS
*****************************************************************************/

/*****************************************************************************
   TYPEDEFS
*****************************************************************************/

/*****************************************************************************
   SYMBOLIC CONSTANTS
*****************************************************************************/

/*****************************************************************************
   RETURN CODES
*****************************************************************************/

/*****************************************************************************
   FUNCTIONS
*****************************************************************************/

#define ObjFusn_START_SEC_SLOW_CODE

/**
 * @brief Initializes the kalman filter track list
 * 
 * Only called once during startup
 * 
 * @return None
*/
void  lkfTrackManagement_init(void);

/**
 * @brief Prepares the Kalman filter for the current cycle
 * 
 * This function is called in pre-cycle function and reduces
 * active track lifespans resulting in cleanup if they are
 * not updated. Also, increases tracks measurement latency
 * by \p f32CycleDt value. It must be called at the beginning
 * of each cycle.
 * 
 * @param[in] f32CycleDt fusion cycle delta time
 * 
 * @return None
*/
void lkfTrackManagement_startCycle(const float32 f32CycleDt);

/**
 * @brief Final step of fusion cycle
 * 
 * This function has to performed once after each fusion cycle. 
 * The lifespan of objects is reduced and if some objects die, they will be dropped.
 * 
 * @param[in] f32CycleDt fusion cycle delta time
 * 
 * @return Error status
*/
uint32 lkfTrackManagement_finalizeCycle(const float32 f32CycleDt);

/**
* @brief   Remove current sensor pattern from track last sensor pattern
*
* \p u32SensorUpdatePattern shows the last sensor's pattern that has provided
* a measurement and was used during association. This function removes this pattern
* from the \p u32SensorsCurr field of tracks that represents the last sensor pattern
* that has detected this track. If the track gets an update, the bit will be set 
* again, otherwise the track can die as soon as the \p u32SensorsCurr equals zero.
* Also, the number of cycles that camera and Radar have not seen each track is updated 
* and track \p au16SensorID is set.  
*
* @param[in] u32SensorUpdatePattern sensor pattern
*
* @return  #TUEOBJFUSN_ERROR_NOERROR if no error occured and error code otherwise.
*/
uint32 lkfTrackManagement_deleteSensorInformation(const uint32 u32SensorUpdatePattern);

/**
 * @brief Predict all internal tracks with given prediction delta time
 * 
 * @param[in] f32PredictionDt prediction delta time [s]
 * 
 * @return error status
 */
uint32 lkfTrackManagement_predictInternalTrkbl(const float32 f32PredictionDt);

/**
 * @brief compensate all internal tracks (TRK) due to ego motion
 * displacement to sensor measurement time
 * 
 * @param[in] f32MeasurementLatency sensor measurement latency
 * 
 * @return error status
 */
extern uint32 lkfTrackManagement_compensateInternalTrkbleList(const float32 f32MeasurementLatency);

/**
 * @brief Update internal track list from association algorithm results
 * 
 * For each track matching that was detected in association algorithm, 
 * internal track is updated with sensor measurement tracks by using
 * Kalman filter update step.
 * 
 * @param[in] pMeasList sensor measurement list
 * @param[in] pMatchIndexArray data association matching index matrix
 * @param[in] pEgoMotion corresponding ego motion dynamics at current time
 * @param[out] abObjectAssignedFlags specifies whether the measurement object 
 * is used for updating associated internal track
 * 
 * @return error status
*/
uint32 lkfTrackManagement_updateTrkList(CONSTP2VAR(TueObjFusn_ObjectListType, AUTOMATIC, ObjFusn_VAR_NOINIT) pMeasList,
                                        CONSTP2CONST(stMatchIndexArrayType, AUTOMATIC, ObjFusn_VAR_NOINIT) pMatchIndexArray, 
                                        VAR(boolean, ObjFusn_VAR_NOINIT) abObjectAssignedFlags[],
                                        CONSTP2CONST(TueObjFusn_EgoMotionType, AUTOMATIC, ObjFusn_VAR_NOINIT) pEgoMotion);

/**
 * @brief setup new tracks from those that are not associated
 * 
 * If there are measurement tracks that are not associated with any tracks
 * during association algorithm, create new tracks from them.
 * 
 * @param[in] pMeasList input sensor track list
 * @param[in] abObjectAssignedFlags boolean array specifying associated tracks
 * 
 * @return error status
*/
uint32 lkfTrackManagement_setupNewLkfTrkble(CONSTP2CONST(TueObjFusn_ObjectListType, AUTOMATIC, ObjFusn_VAR_NOINIT) pMeasList,
       CONST(boolean, ObjFusn_VAR_NOINIT) abObjectAssignedFlags[]);


/**
 * @brief Copy a subset of internal tracks (TRK) to \p pOutputList
 * 
 * Iterates over all active internal tracks (TRK) and only copies
 * those that are not specified as dropped. Then copies them to 
 * \p pOutputList.
 * 
 * @param[in] abObjectsToBeDropped boolean array specifying which objects 
 * will be dropped and should not be copied
 * @param[out] pOutputList pointer to trackable list that is filled
 * 
 * @return error status
*/
uint32 lkfTrackManagement_copyLkfTrkbl2Outputlist(CONSTP2VAR(TueObjFusn_ObjectListType, AUTOMATIC, ObjFusn_VAR_NOINIT) pOutputList,
                                                  const boolean abObjectsToBeDropped[]);

/**
 * @brief Get the age of the track list
 * 
 * Shows that the data was valid (updated/predicted) to lkfTrackManagement_getTrkListAge() sec ago
 * 
 * @return tracks age
*/
extern float32 lkfTrackManagement_getTrkListAge(void);

/**
 * @brief returns the internal track list pointer
 * 
 * This data must be used for reading (not modified)
 * 
 * @return Pointer to track list
*/
extern P2VAR(TueObjFusn_TrackableListType, AUTOMATIC, ObjFusn_CODE) lkfTrackManagement_getTrkbleList(void);

/** 
 * @brief update the object existence probability for the internal track list 
 * 
 * @param[in] u32SensorUpdatePattern list of sensor patterns that
 * have been updated
 * 
 * @return error status
*/
uint32 lkfTrackManagement_updateExistenceProbability(const uint32 u32SensorUpdatePattern);

/**
 * @brief Removes extra tracks that could not be kept due to buffer size
 * 
 * Fusion code can only keep at most #TUEOBJFUSN_TRACKABLELIST_U16VALIDTRACKABLES_MAX tracks
 * (248 currently). Therefore, if there are more tracks than the buffer size (after association),
 * then extra internal tracks are deleted. Deletion priority and order is determined in object selection.
 * Overall, it selects VIO trackables from the internal trackable list if coasting is active.
 * 
 * @param[in] u16NumMeasurements number of sensor measuremnt
 * @param[in] u16nBestMatches number of object matches in data association
 * 
 * @return error status
*/
uint32  lkfTrackManagement_selectVioTrackables(const uint16 u16NumMeasurements, const uint16 u16nBestMatches);

/**
 * @brief Runs the track merge algorithm 
 * 
 * Main function for the track merge algorithm which
 * calls subsequent functions accordingly.
 * 
 * @param[in] pEgoMotion pointer to ego vehicle dynamics at
 * the TRK list age
 * 
 * @return Error status
*/
uint32 lkfTrackManagement_runTrackMerge(CONSTP2CONST(TueObjFusn_EgoMotionType, AUTOMATIC, ObjFusn_VAR_NOINIT) pEgoMotion);

/**
 * @brief increase diagonal elements of covariance matrix (P) by the defined constant
 * 
 * Values are increased by #TUE_PRV_LKF_TRACK_MANAGEMENT_ADDITIONAL_Q_ON_DIAGONAL
 * 
 * @param[in,out] matP covariance matrix
 *
 * @return None
*/
void  lkfTrackManagement_increaseDiagonalP(CONSTP2VAR(stf32SymMatrix_t, AUTOMATIC, ObjFusn_VAR_NOINIT) matP);

/**
 * @brief Remove tracks of the sensor that its timeout is reached
 * 
 * If sensor timeout is reached and it hasn't provided any new data, 
 * non-fused tracks that were detected by this sensor are obsolete and must
 * be deleted. #TUE_PRV_VEHICLE_SENSOR_TIMEOUT defines the timeout value.
 * 
 * @param[in] u32SensorPattern obsolete sensor pattern
 * 
 * @return error status
*/
uint32 lkfTrackManagement_removeObsoleteTracks(const uint32 u32SensorPattern);
#define ObjFusn_STOP_SEC_SLOW_CODE


/*==================[external function declarations]========================*/
/*==================[external constants]====================================*/
/*==================[external data]=========================================*/

#ifdef __cplusplus
}
#endif
#endif /* TUE_PRV_LKF_TRACKMANAGEMENT_H */
/*==================[end of file]===========================================*/
/** @} */
